/**
 * @file bitbang.c
 * Bitbanging over RS-232 lines for POSIX termios and for Microsoft Windows.
 * Written in 2001, 2002, 2004, 2005 by Marko Mkel
 * (http://www.iki.fi/~msmakela/).
 * This file is in the public domain.
 *
 * This program was originally written to prove that the Windows
 * driver of the FTDI FT232 USB-to-serial converter is buggy.  If you
 * physically connect RTS and CTS and change the state of RTS with
 * this program, the driver will report the previous state of the CTS
 * signal.  The manufacturer (support@ftdi.com) ignored my bug report.
 *
 * Note that the CTS bug is not the only bug in the FTDI driver.  Also
 * the Linux driver has bugs, most notably it ignores XON/XOFF
 * flow control.  In fact, the drivers of many USB-to-serial adapters
 * seem to be buggy in one way or another, especially regarding flow
 * control and timeouts.
 */

#if defined __unix || defined __unix__ || defined __APPLE__
# include <termios.h>
# include <unistd.h>
# include <sys/ioctl.h>
# include <fcntl.h>
#endif /* defined __unix || defined __unix__ || defined __APPLE__ */

#if defined __WIN32 || defined __WIN32__
# include <windows.h>
# ifndef WIN32
#  define WIN32
# endif /* WIN32 */
#endif /* defined __WIN32 || defined __WIN32__ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#ifdef WIN32
/** file handle for the serial device */
static HANDLE serialfd = INVALID_HANDLE_VALUE;
/** previous terminal input/output settings for the serial line */
static DCB tio;
/** new terminal input/output settings for the serial line */
static DCB newtio;
#else /* WIN32 */
/** file descriptor for the serial device (-1=none) */
static int serialfd = -1;
/** previous terminal input/output settings for the serial line */
static struct termios tio;
/** new terminal input/output settings for the serial line */
static struct termios newtio;
/** terminal I/O control flags */
static int tiocm;
#endif /* WIN32 */

#ifdef WIN32
/** Alter the state of a communication line
 * @param context	context for the error message
 * @param line		the line to be altered
 * @return		zero on success; nonzero on error
 */
static int
set_line (const char* context,
	  DWORD line)
{
  if (EscapeCommFunction (serialfd, line))
    return 0;
  fprintf (stderr, "%s: EscapeCommFunction failed with code %ld\n",
	   context, GetLastError ());
  return 1;
}

#else /* WIN32 */
/** Alter the state of the communication lines
 * @param context	context for the error message
 * @return		zero on success; nonzero on error
 */
static int
set_lines (const char* context)
{
  if (0 > ioctl (serialfd, TIOCMSET, &tiocm)) {
    fputs (context, stderr);
    perror (": ioctl");
    return 1;
  }
  return 0;
}
#endif /* WIN32 */

/** Open the communication channel
 * @param dev	name of the communication device
 * @return	zero on success, nonzero on failure
 */
int
serial_init (const char* dev)
{
#ifdef WIN32
  if ((serialfd = CreateFile (dev, GENERIC_WRITE | GENERIC_READ, 0, 0,
			      OPEN_EXISTING, 0, 0)) == INVALID_HANDLE_VALUE) {
    fprintf (stderr, "serial_init %s: CreateFile failed with code %ld\n",
	     dev, GetLastError ());
    return 1;
  }
  if (!GetCommState (serialfd, &tio)) {
    fprintf (stderr, "serial_init %s: GetCommState failed with code %ld\n",
	     dev, GetLastError ());
  init_fail:
    CloseHandle (serialfd);
    serialfd = INVALID_HANDLE_VALUE;
    return 2;
  }
  PurgeComm (serialfd, PURGE_TXCLEAR | PURGE_RXCLEAR);
#else /* WIN32 */
  if ((serialfd = open (dev, O_RDWR | O_NOCTTY | O_NONBLOCK)) < 0) {
    fputs ("serial_init ", stderr);
    fputs (dev, stderr);
    perror (": open");
    return 1;
  }
  if (tcgetattr (serialfd, &tio)) {
    fputs ("serial_init ", stderr);
    fputs (dev, stderr);
    perror (": tcgetattr");
  init_fail:
    close (serialfd);
    serialfd = -1;
    return 2;
  }
  /* flush the input and output buffers */
  tcflush (serialfd, TCIOFLUSH);
  if (0 > ioctl (serialfd, TIOCMGET, &tiocm)) {
    fputs ("serial_init ", stderr);
    fputs (dev, stderr);
    perror (": ioctl");
    return 1;
  }
#endif /* WIN32 */

  /* set the communication parameters */
  memcpy (&newtio, &tio, sizeof newtio);
#ifdef WIN32
  newtio.fBinary = 1;
  newtio.fParity = newtio.fOutxCtsFlow = newtio.fOutxDsrFlow = 0;
  newtio.fDtrControl = DTR_CONTROL_DISABLE;
  newtio.fDsrSensitivity = 0;
  newtio.fOutX = 0, newtio.fInX = 0;
  newtio.fErrorChar = newtio.fNull = 0;
  newtio.fRtsControl = RTS_CONTROL_DISABLE;
  newtio.fAbortOnError = 0;
  newtio.ByteSize = 8;
  newtio.Parity = NOPARITY;
  newtio.StopBits = ONESTOPBIT;
  newtio.XonChar = newtio.XoffChar = 0;

  PurgeComm (serialfd, PURGE_TXCLEAR | PURGE_RXCLEAR);
  if (!SetCommState (serialfd, &newtio)) {
    fprintf (stderr, "serial_init %s: SetCommState failed with code %ld\n",
	     dev, GetLastError ());
    goto init_fail;
  }
#else /* WIN32 */
  newtio.c_iflag = IGNBRK | IGNPAR;
  newtio.c_oflag = 0;
  newtio.c_cflag = CS8 | CLOCAL | CREAD;
  newtio.c_lflag = 0;
  memset (newtio.c_cc, 0, sizeof newtio.c_cc);
  newtio.c_cc[VMIN] = 1; /* read at least 1 char */
  newtio.c_cc[VTIME] = 0; /* no timeout */

  /* flush the input and output buffers */
  tcflush (serialfd, TCIOFLUSH);

  if (tcsetattr (serialfd, TCSANOW, &newtio)) {
    fputs ("serial_init ", stderr);
    fputs (dev, stderr);
    perror (": tcsetattr");
    goto init_fail;
  }
#endif /* WIN32 */

  return 0;
}

/** Close the communication channel */
void
serial_fini (void)
{
#ifdef WIN32
  if (serialfd && serialfd != INVALID_HANDLE_VALUE) {
    SetCommState (serialfd, &tio);
    CloseHandle (serialfd);
    serialfd = INVALID_HANDLE_VALUE;
  }
#else /* WIN32 */
  if (serialfd > 0) {
    tcsetattr (serialfd, TCSANOW, &tio);
    close (serialfd);
    serialfd = -1;
  }
#endif /* WIN32 */
}

/** Set DTR
 * @param bit	new state of DTR: zero=negated, nonzero=asserted
 * @return	zero on success, nonzero on failure
 */
int
serial_dtr (int bit)
{
#ifdef WIN32
  return set_line ("serial_dtr", bit ? SETDTR : CLRDTR);
#else /* WIN32 */
  if (bit) tiocm |= TIOCM_DTR; else tiocm &= ~TIOCM_DTR;
  return set_lines ("serial_dtr");
#endif /* WIN32 */
}

/** Set RTS
 * @param bit	new state of RTS: zero=negated, nonzero=asserted
 * @return	zero on success, nonzero on failure
 */
int
serial_rts (int bit)
{
#ifdef WIN32
  return set_line ("serial_rts", bit ? SETRTS : CLRRTS);
#else /* WIN32 */
  if (bit) tiocm |= TIOCM_RTS; else tiocm &= ~TIOCM_RTS;
  return set_lines ("serial_rts");
#endif /* WIN32 */
}

/** Set TXD
 * @param bit	new state of TXD: zero=negated, nonzero=asserted
 * @return	zero on success, nonzero on failure
 */
int
serial_txd (int bit)
{
#ifdef WIN32
  return set_line ("serial_txd", bit ? SETBREAK : CLRBREAK);
#else /* WIN32 */
  if (!ioctl (serialfd, bit ? TIOCSBRK : TIOCCBRK, 0))
    return 0;
  perror ("serial_rts: ioctl");
  return 1;
#endif /* WIN32 */
}

/** Read the state of the communication lines
 * @param cts		out: cts (0=negated, 1=asserted)
 * @param dsr		out: dsr (0=negated, 1=asserted)
 * @return		zero on success; nonzero on error
 */
static int
serial_input (int* cts, int* dsr)
{
#ifdef WIN32
  DWORD lines;
  if (!GetCommModemStatus (serialfd, &lines)) {
    fprintf (stderr,
	     "serial_input: GetCommModemStatus failed with code %ld\n",
	     GetLastError ());
    return -1;
  }
  *cts = !!(lines & MS_CTS_ON);
  *dsr = !!(lines & MS_DSR_ON);
#else /* WIN32 */
  if (0 > ioctl (serialfd, TIOCMGET, &tiocm)) {
    perror ("serial_input: ioctl");
    return 1;
  }
  *cts = !!(tiocm & TIOCM_CTS);
  *dsr = !!(tiocm & TIOCM_DSR);
#endif /* WIN32 */
  return 0;
}

/** The main program
 * @param argc	argument count
 * @param argv	argument vector
 */
int
main (int argc, char** argv)
{
  if (argc != 2) {
    fputs ("Usage: bitbang device, then input [dDrRtT]*\n", stderr);
    return 1;
  }
  if (serial_init (argv[1])) {
    fprintf (stderr, "Failed to initialize serial device %s\n", argv[1]);
    return 2;
  }
  for (;;) {
    int cts, dsr;
    int c = getchar ();
    switch (c) {
    case EOF:
      return 0;
    case 'd': case 'D':
      if (serial_dtr (c & 0x20)) return 3; break;
    case 'r': case 'R':
      if (serial_rts (c & 0x20)) return 3; break;
    case 't': case 'T':
      if (serial_txd (c & 0x20)) return 3; break;
    }
    serial_input (&cts, &dsr);
    fprintf (stderr, "CTS=%d DSR=%d\n", cts, dsr);
  }
  serial_fini ();
  return 0;
}
